use std::rc::{Rc, Weak};
use std::cell::RefCell;
use std::thread::sleep;
use std::time::{Duration, SystemTime};
use r9_macro::SharedFromSelf;
use r9_macro_derive::SharedFromSelf;

pub trait UserApp {
    fn get_pkg_name(&self) -> String;
    fn init(&mut self);
    fn update(&mut self);
    fn uninit(&mut self);
}

#[derive(SharedFromSelf)]
pub struct App {
    _self_wp: Weak::<RefCell::<Self>>,
    zone_id: i32,
    node_id: i32,
    instance_id: i32,
    user_app: Option<Rc::<RefCell::<dyn UserApp>>>,
}

impl App {

    pub fn instance() -> Rc::<RefCell::<Self>> {
        static mut _INSTANCE: Option<Rc::<RefCell::<App>>> = None;
        unsafe {
            match &_INSTANCE {
                Some(v) => {
                    return v.clone();
                }
                None => {
                    _INSTANCE = Some(Rc::new(RefCell::new(
                        App{
                            zone_id: 0,
                            node_id: 0,
                            instance_id: 0,
                            user_app: None,
                            _self_wp: Default::default(),
                        }
                    )));
                    _INSTANCE.clone().unwrap().borrow_mut()._self_wp =
                        Rc::downgrade(&_INSTANCE.clone().unwrap());
                }
            }
            return _INSTANCE.clone().unwrap().clone();
        }
    }

    pub fn init(&mut self, user_app: Rc::<RefCell::<dyn UserApp>>) {
        self.user_app = Some(user_app);
        crate::Timer::instance().borrow_mut().init();
        self.user_app.as_ref().unwrap().borrow_mut().init();
    }

    pub fn uninit(&mut self) {
        self.user_app.as_ref().unwrap().borrow_mut().uninit();
        crate::Timer::instance().borrow_mut().uninit();
    }

    pub fn run(&mut self) {
        while true {
            crate::Timer::instance().borrow_mut().update();
            sleep(Duration::from_secs(1));
        }
    }

    pub fn get_pkg_name(&self) -> String {
        return self.user_app.as_ref().unwrap().borrow().get_pkg_name();
    }

    pub fn new_uuid(&self) -> i64 {
        return 0;
    }

    pub fn get_instance_id(&self) -> i32 {
        return 0;
    }

    pub fn get_zone_id(&self) -> i32 {
        return 0;
    }

    pub fn get_node_id(&self) -> i32 {
        return 0;
    }

    pub fn has_flag(&self) -> bool {
        return false;
    }

    pub fn get_now_second(&self) -> i64 {
        return 0;
    }

    pub fn get_now_millis(&self) -> i64 {
        return 0;
    }

    pub fn get_time_offset(&self) -> i64 {
        return 0;
    }

}
